///
/// Copyright (c) 2015, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.hdict_builder;

import 'HGrid.dart';
import 'HVal.dart';
import 'HStr.dart';
import 'HNum.dart';
import 'HBool.dart';
import 'HMarker.dart';
import 'UnknownNameError.dart';

/// HDictBuilder is used to construct an immutable HDict instance.
///
/// See [Project Haystack](http://project-haystack.org/doc/TagModel#tagKinds)
class HDictBuilder {

  Map<String, HVal> _map = new Map<String, HVal>();

  HDictBuilder();

  /** Add tag name and value.  Return this. */
  HDictBuilder add(String name, [Object val, String unit]) {
    if (!HDict.isTagName(name))
      throw new ArgumentError("Invalid tag name: $name");

    if (val == null) {
      _add(name, HMarker.VAL);
    } else if (val is bool) {
      _add(name, new HBool(val));
    } else if (val is num) {
      _add(name, new HNum(val, unit));
    } else if (val is String) {
      _add(name, new HStr(val));
    } else if (val is HVal) {
      _add(name, val);
    }

    return this;
  }

  /** Add all the name/value pairs in given HDict.  Return this. */
  void addDict(HDict dict) {
    for (Iterator<DictEntry> it = dict.iterator(); it.moveNext(); ) {
      _add(it.current.key, it.current.val);
    }
  }

  void _add(String name, HVal val) {
    _map[name] = val;
  }

  /** Convert current state to an immutable HDict instance */
  HDict toDict() {
    if (_map.isEmpty) return HDict.EMPTY;
    HDict dict = new HDict(this._map);
    this._map = new Map<String, HVal>();
    return dict;
  }

  //////////////////////////////////////////////////////////////////////////
  // Access
  //////////////////////////////////////////////////////////////////////////

  /** Return if size is zero */
  bool isEmpty() => size == 0;

  /** Return number of tag name/value pairs */
  int get size =>  _map.length;

  /** Return if the given tag is present */
  bool has(String name) => get(name, false) != null;

  /** Return if the given tag is not present */
  bool missing(String name) => get(name, false) == null;

  /** Get a tag by name.  If not found and checked if false then
      return null, otherwise throw UnknownNameException */
  HVal get(String name, [bool checked = true]) {
    HVal val = _map[name];
    if (val != null) return val;
    if (!checked) return null;
    throw new UnknownNameError(name);
  }

}