///
/// Copyright (c) 2015, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.hproj;

import 'dart:async';

import 'HRef.dart';
import 'HGrid.dart';
import 'HNum.dart';
import 'HWatch.dart';
import 'UnknownRecError.dart';

/// HProj is the common interface for HClient and HServer to provide
/// access to a database tagged entity records.
///
/// See [Project Haystack](http://project-haystack.org/doc/TagModel)
abstract class HProj {

//////////////////////////////////////////////////////////////////////////
// Operations
//////////////////////////////////////////////////////////////////////////

  /**
   * Get the summary "about" information.
   */
  Future<HDict> about();


//////////////////////////////////////////////////////////////////////////
// Read by id
//////////////////////////////////////////////////////////////////////////

  /**
   * Call "read" to lookup an entity record by it's unique identifier.
   * If not found then return null or throw an UnknownRecException based
   * on checked.
   */
  Future<HDict> readById(HRef id, {bool checked: true}) {
    return onReadById(id).then((HDict rec) {
      if (rec != null) return rec;
      if (checked) throw new UnknownRecError(id: id);
      return null;
    });
  }

  /**
   * Read a list of entity records by their unique identifier.
   * Return a grid where each row of the grid maps to the respective
   * id array (indexes line up).  If checked is true and any one of the
   * ids cannot be resolved then raise UnknownRecException for first id
   * not resolved.  If checked is false, then each id not found has a
   * row where every cell is null.
   */
  Future<HGrid> readByIds(List<HRef> ids, {bool checked: true}) {
    return onReadByIds(ids).then((HGrid grid) {
      if (checked) {
        for (int i=0; i<grid.numRows; ++i)
          if (grid.row(i).missing("id")) throw new UnknownRecError(id: ids[i]);
      }
      return grid;
    });
  }

  /**
   * Subclass hook for readById, return null if not found.
   * @protected
   */
  Future<HDict> onReadById(HRef id);

  /**
   * Subclass hook for readByIds, return rows with nulls cells
   * for each id not found.
   * @protected
   */
  Future<HGrid> onReadByIds(List<HRef> ids);

//////////////////////////////////////////////////////////////////////////
// Read by filter
//////////////////////////////////////////////////////////////////////////

  /**
   * Query one entity record that matches the given filter.  If
   * there is more than one record, then it is undefined which one is
   * returned.  If there are no matches than return null or raise
   * UnknownRecException based on checked flag.
   */
  Future<HDict> read(String filter, {bool checked:true}) {
    return readAll(filter, limit:1).then((HGrid grid) {
      if (grid.numRows > 0) return grid.row(0);
      if (checked) throw new UnknownRecError(msg: filter);
      return null;
    });
  }

  /**
   * Call "read" to query every entity record that matches given filter.
   * Clip number of results by "limit" parameter. Default is 2147483647 - max 32 bit int value.
   */
  Future<HGrid> readAll(String filter, {int limit:2147483647}) => onReadAll(filter, limit);

  /**
   * Subclass hook for read and readAll.
   * @protected
   */
  Future<HGrid> onReadAll(String filter, int limit);


//////////////////////////////////////////////////////////////////////////
// Watches
//////////////////////////////////////////////////////////////////////////

  /**
   * Create a new watch with an empty subscriber list.  The dis
   * string is a debug string to keep track of who created the watch.
   * Pass the desired lease time or null to use default.
   */
  HWatch watchOpen(String dis, HNum lease);

  /**
   * List the open watches.
   */
  List<HWatch> watches();

  /**
   * Lookup a watch by its unique identifier.  If not found then
   * raise UnknownWatchErr or return null based on checked flag.
   */
  HWatch watch(String id, {bool checked:true});

//////////////////////////////////////////////////////////////////////////
// Historian
//////////////////////////////////////////////////////////////////////////

//  /**
//   * Read history time-series data for given record and time range. The
//   * items returned are exclusive of start time and inclusive of end time.
//   * Raise exception if id does not map to a record with the required tags
//   * "his" or "tz".  The range may be either a String or a HDateTimeRange.
//   * If HTimeDateRange is passed then must match the timezone configured on
//   * the history record.  Otherwise if a String is passed, it is resolved
//   * relative to the history record's timezone.
//   */
//  Future<HGrid> hisRead(HRef id, Object range);
//
//  /**
//   * Write a set of history time-series data to the given point record.
//   * The record must already be defined and must be properly tagged as
//   * a historized point.  The timestamp timezone must exactly match the
//   * point's configured "tz" tag.  If duplicate or out-of-order items are
//   * inserted then they must be gracefully merged.
//   */
//  void hisWrite (HRef id, List<HHisItem> items);
}