///
/// Copyright (c) 2015, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.src.hclient;
import 'dart:async';
import 'dart:html' show HttpRequest;

import '../HProj.dart';
import '../HVal.dart';
import '../HRef.dart';
import '../HStr.dart';
import '../HNum.dart';
import '../HGrid.dart';
import '../HWatch.dart';
import '../HGridBuilder.dart';
import '../HDictBuilder.dart';
import '../UnknownWatchError.dart';
import '../UnknownRecError.dart';
import '../io/HZincWriter.dart';
import '../io/HZincReader.dart';

part 'CallErrError.dart';
part 'CallError.dart';

/// HClient manages a logical connection to a HTTP REST haystack server.
///
/// See [Project Haystack](http://project-haystack.org/doc/Rest)
class HClient extends HProj {

  factory HClient(String uri, String user, String pass) {
    // check uri
//    if (!uri.startsWith("http://") && !uri.startsWith("https://")) throw new ArgumentError("Invalid uri format: " + uri);
    if (!uri.endsWith("/")) uri = uri + "/";

    // sanity check arguments
    if (user.length == 0) throw new ArgumentError("user cannot be empty string");

    return new HClient._(uri, user, pass);
  }

  HClient._(this.uri, this._user, this._pass);

//////////////////////////////////////////////////////////////////////////
// State
//////////////////////////////////////////////////////////////////////////

  /** Base URI for connection such as "http://host/api/demo/".
      This string always ends with slash. */
  final String uri;

  /** Timeout in milliseconds for opening the HTTP socket */
  int connectTimeout = 60 * 1000;

  /** Timeout in milliseconds for reading from the HTTP socket */
  int readTimeout = 60 * 1000;

//////////////////////////////////////////////////////////////////////////
// Operations
//////////////////////////////////////////////////////////////////////////

  /**
   * Authenticate the client and return this.
   */
  Future<HClient> open() {
    // fixme need to implement auth
    // authenticate();
    return new Future(() => this);
  }

  /**
   * Call "about" to query summary info.
   */
  Future<HDict> about() => call("about", HGrid.EMPTY).then((HGrid res) => res.row(0));

  /**
   * Call "ops" to query which operations are supported by server.
   */
  Future<HGrid> ops() {
    return call("ops", HGrid.EMPTY);
  }

  /**
   * Call "formats" to query which MIME formats are available.
   */
  Future<HGrid> formats() {
    return call("formats", HGrid.EMPTY);
  }


//////////////////////////////////////////////////////////////////////////
// Reads
//////////////////////////////////////////////////////////////////////////

  Future<HDict> onReadById(HRef id) {
    return readByIds(new List<HRef>.from([id]), checked:false).then((HGrid res) {
      if (res.isEmpty) return null;
      HDict rec = res.row(0);
      if (rec.missing("id")) return null;
      return rec;
    });
  }

  Future<HGrid> onReadByIds(List<HRef> ids) {
    HGridBuilder b = new HGridBuilder();
    b.addCol("id");

    for (int i=0; i<ids.length; ++i)
      b.addRow(new List<HVal>.from([ids[i]]));

    HGrid req = b.toGrid();
    return call("read", req);
  }

  Future<HGrid> onReadAll(String filter, int limit) {
    HGridBuilder b = new HGridBuilder();
    b.addCol("filter");
    b.addCol("limit");
    b.addRow(new List<HVal>.from([new HStr(filter), new HNum(limit)]));
    HGrid req = b.toGrid();
    return call("read", req);
  }

//////////////////////////////////////////////////////////////////////////
// Evals
//////////////////////////////////////////////////////////////////////////

  /**
   * Call "eval" operation to evaluate a vendor specific
   * expression on the server:
   *   - SkySpark: any Axon expression
   *
   * Raise CallErrException if the server raises an exception.
   */
  Future<HGrid> eval(String expr) {
    HGridBuilder b = new HGridBuilder();
    b.addCol("expr");
    b.addRow([new HStr(expr)]);
    HGrid req = b.toGrid();
    return call("eval", req);
  }

  /**
   * Convenience for "evalAll(HGrid, checked)".
   */
  Future<List<HGrid>> evalAllStrings(List<String> exprs, {bool checked:true}) {
    HGridBuilder b = new HGridBuilder();
    b.addCol("expr");

    for (int i=0; i<exprs.length; ++i)
      b.addRow([ new HStr(exprs[i]) ]);

    return evalAll(b.toGrid(), checked:checked);
  }

  /**
   * Call "evalAll" operation to evaluate a batch of vendor specific
   * expressions on the server. See "eval" method for list of vendor
   * expression formats.  The request grid must specify an "expr" column.
   * A separate grid is returned for each row in the request.  If checked
   * is false, then this call does *not* automatically check for error
   * grids.  Client code must individual check each grid for partial
   * failures using "Grid.isErr".  If checked is true and one of the
   * requests failed, then raise CallErrException for first failure.
   */
  Future<List<HGrid>> evalAll(HGrid req, { bool checked: true}) {
    String reqStr = HZincWriter.gridToString(req);

    return _postString("evalAll", reqStr).then((String resStr) {
      List<HGrid> res = new HZincReader(resStr).readGrids();
      if (checked) {
        for (int i=0; i<res.length; ++i)
          if (res[i].isErr) throw new CallErrError(res[i]);
      }

      return res;
    });
  }

//////////////////////////////////////////////////////////////////////////
// Watches
//////////////////////////////////////////////////////////////////////////

  /**
   * Create a new watch with an empty subscriber list.  The dis
   * string is a debug string to keep track of who created the watch.
   */
  HWatch watchOpen(String dis, HNum lease) {
    return new HClientWatch(this, dis, lease);
  }

  /**
   * List the open watches associated with this HClient.
   * This list does *not* contain a watch until it has been successfully
   * subscribed and assigned an identifier by the server.
   */
  List<HWatch> watches() => _watches.values.toList(growable:false);

  /**
   * Lookup a watch by its unique identifier associated with this HClient.
   * If not found return null or raise UnknownWatchException based on
   * checked flag.
   */
  HWatch watch(String id, {bool checked:true}) {
    HWatch w = _watches[id];
    if (w != null) return w;
    if (checked) throw new UnknownWatchError(id);
    return null;
  }


  Future<HGrid> watchSub(HClientWatch w, List<HRef> ids, {bool checked:true}) {
    if (ids.length == 0) throw new ArgumentError("ids are empty");
    if (w.isClosed) throw new StateError("watch is closed");

    // grid meta
    HGridBuilder b = new HGridBuilder();
    if (w.id != null) b.meta.add("watchId", w.id);
    if (w._desiredLease != null) b.meta.add("lease", w._desiredLease);
    b.meta.add("watchDis", w.dis);

    // grid rows
    b.addCol("id");
    for (int i=0; i<ids.length; ++i)
      b.addRow(new List<HVal>.from([ids[i]]));

    // make request
    return call("watchSub", b.toGrid()).then((HGrid res) {
      // make sure watch is stored with its watch id
      if (w.id == null) {
        w._id = res.meta.getStr("watchId");
        w._lease = res.meta.get("lease") as HNum;
        _watches[w.id] = w;
      }

      // if checked, then check it
      if (checked) {
        if (res.numRows != ids.length && ids.length > 0)
          throw new UnknownRecError(id: ids[0]);
        for (int i=0; i<res.numRows; ++i)
          if (res.row(i).missing("id")) throw new UnknownRecError(id: ids[i]);
      }

      return res;
    },
    onError: (e) {
      if (e is CallErrError) {
        // any server side error is considered close
        watchClose(w, send: false);
      }

      throw e;
    });
  }


  Future<HGrid> watchUnsub(HClientWatch w, List<HRef> ids) {
    if (ids.length == 0) throw new ArgumentError("ids are empty");
    if (w.id == null) throw new StateError("nothing subscribed yet");
    if (w.isClosed) throw new StateError("watch is closed");

    // grid meta
    HGridBuilder b = new HGridBuilder();
    b.meta.add("watchId", w.id);

    // grid rows
    b.addCol("id");
    for (int i=0; i<ids.length; ++i)
    b.addRow([ids[i]]);

    // make request
    return call("watchUnsub", b.toGrid());
  }


  Future<HGrid> watchPoll(HClientWatch w, {bool refresh:false}) {
    if (w.id == null) throw new StateError("nothing subscribed yet");
    if (w.isClosed) throw new StateError("watch is closed");

    // grid meta
    HGridBuilder b = new HGridBuilder();
    b.meta.add("watchId", w.id);
    if (refresh) b.meta.add("refresh");
    b.addCol("empty");

    // make request
    return call("watchPoll", b.toGrid()).catchError((e) {
      if (e is CallErrError) {
        // any server side error is considered close
        watchClose(w, send: false);
      }
      throw e;
    });
  }


  Future watchClose(HClientWatch w, {bool send:true}) {
    // mark flag on watch itself, short circuit if already closed
    if (w.isClosed) return new Future.value();
    w._closed = true;

    // remove it from my lookup table
    if (w.id != null) _watches.remove(w.id);

    // optionally send close message to server
    if (send) {
      HGridBuilder b = new HGridBuilder();
      b.meta.add("watchId", w.id).add("close");
      b.addCol("id");
      return call("watchUnsub", b.toGrid()).catchError((e){});
    } else {
      return new Future.value();
    }
  }


//////////////////////////////////////////////////////////////////////////
// PointWrite
//////////////////////////////////////////////////////////////////////////

  /**
   * Write to a given level of a writable point, and return the current status
   * of a writable point's priority array (see pointWriteArray()).
   *
   * @param id Ref identifier of writable point
   * @param level Number from 1-17 for level to write
   * @param val value to write or null to auto the level
   * @param who optional username performing the write, otherwise user dis is used
   * @param duration Number with duration unit if setting level 8
   */
  Future<HGrid> pointWrite(
      HRef id, int level, String who,
      HVal val, HNum dur)
  {
    HGridBuilder b = new HGridBuilder();
    b.addCol("id");
    b.addCol("level");
    b.addCol("who");
    b.addCol("val");
    b.addCol("duration");

    b.addRow([
      id,
      new HNum(level),
      new HStr(who),
      val,
      dur
    ]);

    HGrid req = b.toGrid();
    return call("pointWrite", req);
  }

  /**
   * Return the current status
   * of a point's priority array.
   * The result is returned grid with following columns:
   * <ul>
   *   <li>level: number from 1 - 17 (17 is default)
   *   <li>levelDis: human description of level
   *   <li>val: current value at level or null
   *   <li>who: who last controlled the value at this level
   * </ul>
   */
  Future<HGrid> pointWriteArray(HRef id) {
    HGridBuilder b = new HGridBuilder();
    b.addCol("id");
    b.addRow([id]);

    HGrid req = b.toGrid();
    return call("pointWrite", req);
  }


//////////////////////////////////////////////////////////////////////////
// History
//////////////////////////////////////////////////////////////////////////

  /**
   * Read history time-series data for given record and time range. The
   * items returned are exclusive of start time and inclusive of end time.
   * Raise exception if id does not map to a record with the required tags
   * "his" or "tz".  The range may be either a String or a HDateTimeRange.
   * If HTimeDateRange is passed then must match the timezone configured on
   * the history record.  Otherwise if a String is passed, it is resolved
   * relative to the history record's timezone.
   */
  Future<HGrid> hisRead(HRef id, Object range) {
    HGridBuilder b = new HGridBuilder();
    b.addCol("id");
    b.addCol("range");
    b.addRow([id, new HStr(range.toString())]);
    HGrid req = b.toGrid();
    return call("hisRead", req);
  }

  /**
   * Write a set of history time-series data to the given point record.
   * The record must already be defined and must be properly tagged as
   * a historized point.  The timestamp timezone must exactly match the
   * point's configured "tz" tag.  If duplicate or out-of-order items are
   * inserted then they must be gracefully merged.
   */
  Future<HGrid> hisWrite(HRef id, List<HHisItem> items) {
    HDict meta = new HDictBuilder().add("id", id).toDict();
    HGrid req = HGridBuilder.hisItemsToGrid(meta, items);
    return call("hisWrite", req);
  }


//////////////////////////////////////////////////////////////////////////
// Actions
//////////////////////////////////////////////////////////////////////////

  /**
   * Invoke a remote action using the "invokeAction" REST operation.
   */
  Future<HGrid> invokeAction(HRef id, String action, HDict args){
    HDict meta = new HDictBuilder().add("id", id).add("action", action).toDict();
    HGrid req = HGridBuilder.dictsToGrid(meta, [args]);
    return call("invokeAction", req);
  }


//////////////////////////////////////////////////////////////////////////
// Call
//////////////////////////////////////////////////////////////////////////

  /**
   * Make a call to the given operation.  The request grid is posted
   * to the URI "this.uri+op" and the response is parsed as a grid.
   * Raise CallNetworkException if there is a communication I/O error.
   * Raise CallErrException if there is a server side error and an error
   * grid is returned.
   */
  Future<HGrid> call(String op, HGrid req) {
    return _postGrid(op, req).then((HGrid res) {
      if (res.isErr) throw new CallErrError(res);
      return res;
    });
  }

  Future<HGrid> _postGrid(String op, HGrid req) {
    String reqStr = HZincWriter.gridToString(req);
    return _postString(op, reqStr).then((String resStr) => new HZincReader(resStr).readGrid());
  }

  Future<String> _postString(String op, String req) {
    return HttpRequest.request(uri + op, method:'POST', sendData: req).then((HttpRequest request) {
      return request.responseText;
    });
  }


//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  final String _user;
  final String _pass;
//  Property authProperty;
//  Property cookieProperty;

  Map<String, HWatch> _watches = new Map<String, HWatch>();
}

class HClientWatch extends HWatch {

  HClientWatch(HClient this._client, String this._dis, HNum this._desiredLease);

  String get id => _id;
  HNum lease() => _lease;
  String get dis => _dis;
  bool get isOpen => !_closed;
  bool get isClosed => _closed;

  Future<HGrid> sub(List<HRef> ids, {bool checked:true}) => _client.watchSub(this, ids, checked:checked);
  Future<HGrid> unsub(List<HRef> ids) => _client.watchUnsub(this, ids);
  Future<HGrid> pollChanges() => _client.watchPoll(this, refresh:false);
  Future<HGrid> pollRefresh() => _client.watchPoll(this, refresh:true);
  Future close() => _client.watchClose(this, send: true);

  final HClient _client;
  final String _dis;
  final HNum _desiredLease;
  HNum _lease;
  String _id;
  bool _closed = false;

}