///
/// Copyright (c) 2015, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///
@TestOn("browser")
library haystack.client_test;

import 'dart:async';

import 'package:test/test.dart';

import 'package:haystack/core.dart';
import 'package:haystack/hclient.dart';

final String uri = "http://localhost:8080/api/demo";
final String user = "haystack";
final String pass = "testpass";

void main() {
  HClient client;

  group('HClient', () {

    // create proper client
    setUp(() {
      if (client == null) {
        client = new HClient(uri, user, pass);
        return client.open();
      }
    });


    group('About', () {
      HDict about;
      setUp(() {
        return client.about().then((HDict res) {
          about = res;
        });
      });

      test('version, product name', () {
        expect(about.getStr("haystackVersion"), equals("2.0"));
        expect(about.getStr("productName"), equals("SkySpark"));
        // expect(result.getStr("tz"), equals(HTimeZone.DEFAULT.name)); //stupid timezones
      });
    });


    group('Ops', () {
      HGrid ops;
      setUp(() {
        if (ops == null) {
          return client.ops().then((HGrid res) {
            ops = res;
          });
        }
      });

      test('verify required columns', () {
        expect(ops.colByName("name")  != null, equals(true));
        expect(ops.colByName("summary") != null, equals(true));
      });

      test('verify required ops', () {
        verifyGridContains(ops, "name", "about");
        verifyGridContains(ops, "name", "ops");
        verifyGridContains(ops, "name", "formats");
        verifyGridContains(ops, "name", "read");
      });
    });


    group('Formats', () {
      HGrid formats;
      setUp(() {
        if (formats == null) {
          return client.formats().then((HGrid res) {
            formats = res;
          });
        }
      });

      test('verify required columns', () {
        expect(formats.colByName("mime")  != null, equals(true));
        expect(formats.colByName("receive") != null, equals(true));
        expect(formats.colByName("send") != null, equals(true));
      });

      test('verify required formats', () {
        verifyGridContains(formats, "mime", "text/plain");
        verifyGridContains(formats, "mime", "text/zinc");
      });
    });


    group('Read', () {
      String disA = "Gaithersburg";
      String disB = "Carytown";

      HRef badId = new HRef("badBadId");

      HDict recA;
      HDict recB;

      setUp(() {
        Future futA = client.read("site and dis==\"$disA\"").then((HDict value) => recA = value);
        Future futB = client.read("site and dis==\"$disB\"").then((HDict value) => recB = value);

        return Future.wait([futA, futB]);
      });

      test('read', () {
        expect(recA.dis(), equals(disA));
        expect(recB.dis(), equals(disB));

        Future<HDict> badTagDict = client.read("badTagShouldBeThere", checked: false);
        expect(badTagDict, completion(isNull));

        Future<HDict> badTagDict2 = client.read("badTagShouldBeThere");
        expect(badTagDict2, throwsA(new isInstanceOf<UnknownRecError>()));
      });

      HGrid grid;
      setUp(() {
        if (grid == null) {
          return client.readAll('site').then((HGrid res) {
            grid = res;
          });
        }
      });

      test('readAll', () {
        verifyGridContains(grid, "dis", disA);
        verifyGridContains(grid, "dis", disB);
        verifyGridContains2(grid, "id", recA.id());
        verifyGridContains2(grid, "id", recB.id());
      });

      test('readAll limit', () {
        expect(grid.numRows, greaterThan(2));
        Future<int> numRowsF = client.readAll("site", limit: 2).then((HGrid val) => val.numRows);
        expect(numRowsF, completion(equals(2)));
      });

      test('readById', () {
        Future<String> recDis = client.readById(recA.id()).then((HDict val) => val.dis());
        expect(recDis, completion(equals(disA)));

        expect(client.readById(badId, checked: false), completion(isNull));
        expect(client.readById(badId), throwsA(new isInstanceOf<UnknownRecError>()));
      });

      setUp(() => client.readByIds([recA.id(), recB.id()])
      .then((HGrid res) => grid = res)
      );

      test('readByIds', () {
        expect(grid.numRows, equals(2));
        expect(grid.row(0).dis(), equals(disA));
        expect(grid.row(1).dis(), equals(disB));
      });

      setUp(() => client.readByIds([recA.id(), badId, recB.id()], checked: false)
      .then((HGrid res) => grid = res)
      );

      test('readByIds (bad id)', () {
        expect(grid.numRows, equals(3));
        expect(grid.row(0).dis(), equals(disA));
        expect(grid.row(1).missing('id'), isTrue);
        expect(grid.row(2).dis(), equals(disB));

        expect(client.readByIds([recA.id(), badId]), throwsA(new isInstanceOf<UnknownRecError>()));
      });

    });


    group('Eval', () {
      test('read(today)', () => client.eval("today()").then((HGrid g ) {
        expect(g.row(0).get("val"), equals(HDate.today()));
      }));

      test('readAll(ahu)', () => client.eval("readAll(ahu)").then((HGrid g ) {
        expect(g.numRows, greaterThan(0));
        verifyGridContains(g, "navName", "RTU-1");
      }));

      test('evalAll', () => client.evalAllStrings(["today()", "[10, 20, 30]", "readAll(site)"]).then((List<HGrid> grids) {
        expect(grids.length, equals(3));

        HGrid g = grids[0];
        expect(g.numRows, equals(1));
        expect(g.row(0).get("val"), equals(HDate.today()));

        g = grids[1];
        expect(g.numRows, equals(3));
        expect(g.row(0).get("val"), new HNum(10));
        expect(g.row(1).get("val"), new HNum(20));
        expect(g.row(2).get("val"), new HNum(30));

        g = grids[2];
        expect(g.numRows, greaterThan(2));
        verifyGridContains(g, "dis", "Carytown");
      }));

      test('evalAll (bad id)', () => client.evalAllStrings(["today()", "readById(@badBadBadId)"], checked:false).then((List<HGrid> grids) {
        expect(grids.length, equals(2));
        expect(grids[0].isErr, isFalse);
        expect(grids[0].row(0).get("val"), equals(HDate.today()));
        expect(grids[1].isErr, isTrue);
      }));

      test('evalAll (bad id, exception)', () => expect(
          client.evalAllStrings(["today()", "readById(@badBadBadId)"]), throwsA(new isInstanceOf<CallErrError>())
      ));
    });


    group('Watchers', () {
      final String watchDis = "Java Haystack Test " + new DateTime.now().millisecondsSinceEpoch.toString();
      HWatch w;

      test('create new watch',() {
        w = client.watchOpen(watchDis, new HNum(3, "min"));
        expect(w.id, isNull);
        expect(w.dis, equals(watchDis));
        expect(w.lease(), isNull);
      });

      HGrid recs;
      HDict a;
      HDict b;
      HDict c;
      HDict d;

      test('do query to get some recs', () => client.readAll('ahu').then((HGrid result) {
        recs = result;
        expect(recs.numRows, greaterThanOrEqualTo(4));
        a = recs.row(0);
        b = recs.row(1);
        c = recs.row(2);
        d = recs.row(3);
      }));

      test('do first sub', () => w.sub([a.id(), b.id()]).then((HGrid sub) {
        expect(sub.numRows, equals(2));
        expect(sub.row(0).dis(), a.dis());
        expect(sub.row(1).dis(), b.dis());
      }));

      test('now add c, bad, d', () => w.sub([c.id(), new HRef('badBadBad'), d.id()], checked:false).then((HGrid sub) {
        expect(sub.numRows, equals(3));
        expect(sub.row(0).dis(), equals(c.dis()));
        expect(sub.row(1).missing("id"), isTrue);
        expect(sub.row(2).dis(), equals(d.dis()));
      }));

      test('now add c, bad, d (with exception)', () {
        expect(w.sub([new HRef('badBadBad')]), throwsA(new isInstanceOf<UnknownRecError>()));
      });

      test('verify state of watch now', () {
        expect(client.watch(w.id), same(w));
        expect(client.watches().length, equals(1));
        expect(client.watches()[0], same(w));
        expect(w.lease().millis(), equals(180000));
      });

      test('poll for changes (should be none yet)', () => w.pollChanges().then((HGrid poll) {
        expect(poll.numRows, equals(0));
      }));

      group('Make change to b and d', () {
        test('Initial check to have javaTest (should be hasn\'t)', () {
          expect(b.has("javaTest"), isFalse);
          expect(d.has("javaTest"), isFalse);
        });

        test('Commit something', () {
          Future fB = client.eval("commit(diff(readById(@" + b.id().val + "), {javaTest:123}))");
          Future fD = client.eval("commit(diff(readById(@" + d.id().val + "), {javaTest:456}))");
          return Future.wait([fB, fD]);
        });

        test('Poll and check', () => w.pollChanges().then((HGrid poll) {
          expect(poll.numRows, equals(2));

          HDict newb, newd;
          if (poll.row(0).id() == b.id()) {
            newb = poll.row(0);
            newd = poll.row(1);
          } else {
            newb = poll.row(1);
            newd = poll.row(0);
          }

          expect(newb.id(), equals(b.id()));
          expect(newd.id(), equals(d.id()));
          expect(newb.get("javaTest"), equals(new HNum(123)));
          expect(newd.get("javaTest"), equals(new HNum(456)));
        }));
      });

      test('poll refresh', () => w.pollRefresh().then((HGrid poll) {
        expect(poll.numRows, equals(4));
        verifyGridContains2(poll, "id", a.id());
        verifyGridContains2(poll, "id", b.id());
        verifyGridContains2(poll, "id", c.id());
        verifyGridContains2(poll, "id", d.id());
      }));

      test('remove d, and then poll changes',
          () => w.unsub([d.id()])
      .then((_) {
        return client.eval("commit(diff(readById(@" + b.id().val + "), {-javaTest}))");
      }).then((_) {
        return client.eval("commit(diff(readById(@" + d.id().val + "), {-javaTest}))");
      }).then((_) {
        return w.pollChanges().then((HGrid poll) {
          expect(poll.numRows, equals(1));
          expect(poll.row(0).id(), equals(b.id()));
          expect(poll.row(0).has("javaTest"), isFalse);
        });
      })
      );

      test('remove a and c and poll refresh', () => w.unsub([a.id(), c.id()])
      .then((_) => w.pollRefresh()
      .then((HGrid poll) {
        expect(poll.numRows, equals(1));
        expect(poll.row(0).id(), equals(b.id()));
      })
      )
      );

      test('close', () {
        String expr = "folioDebugWatches().findAll(x=>x->dis.contains(\"" + watchDis + "\")).size";

        return client.eval(expr).then((HGrid res) {
          expect(res.row(0).getInt("val"), equals(1));
        }).then((_) {
          return w.close();
        }).then((_) {
          expect(() => w.pollRefresh(), throws);
        }).then((_) => client.eval(expr).then((HGrid res) {
          expect(res.row(0).getInt("val"), equals(0));
          expect(client.watch(w.id, checked: false), isNull);
          expect(client.watches().length, equals(0));
        }));
      });
    });


    group('His Reads', () {
      HDateTime ts(HDict r, [String col = 'ts']) => r.get(col);
      HNum numVal(HRow r) => r.get("val");

      HDict kw;
      setUp(() => client.read("power and siteMeter").then((HDict val) => kw = val));

      test('', () => client.hisRead(kw.id(), "yesterday").then((HGrid his) {
        expect(his.meta.id(), kw.id());
        expect(ts(his.meta, "hisStart").date, equals(HDate.today().minusDays(1)));
        expect(ts(his.meta, "hisEnd").date, equals(HDate.today()));
        expect(his.numRows, greaterThan(90));
        int last = his.numRows-1;
        expect(ts(his.row(0)).date, equals(HDate.today().minusDays(1)));
        expect(ts(his.row(0)).time, equals(new HTime(0, 15)));
        expect(ts(his.row(last)).date, equals(HDate.today()));
        expect(ts(his.row(last)).time, equals(new HTime(0, 0)));
        expect(numVal(his.row(0)).unit, equals("kW"));
      }));
    });


    group('His Write', () {
      Future clearHisWrite(HDict rec) {
        // existing data and verify we don't have any data for 7 June 20120
        String expr = "hisClear(@" + rec.id().val + ", 2010-06)";
        return client.eval(expr).then((_) => client.hisRead(rec.id(), "2010-06-07").then((HGrid his) {
          expect(his.numRows, equals(0));
        }));
      }

      HDict kw;
      List<HHisItem> write;

      setUp(() {
        return client.read("power and not siteMeter").then((HDict res) {
          // setup test
          kw = res;
        }).then((_) {
          return clearHisWrite(kw);
        }).then((_) {
          // create some items
          HDate date = new HDate(2010, 6, 7);
          HTimeZone tz = new HTimeZone(kw.getStr("tz")); //fixme here only New_york, without any time shift
          write = new List<HHisItem>(5);
          for (int i=0; i<write.length; ++i) {
            HDateTime ts = new HDateTime(date, new HTime(i+1, 0), tz);
            HVal val = new HNum(i, "kW");
            write[i] = new HHisItem(ts, val);
          }
        });
      });

      // clean test after it's complete
      tearDown(() => clearHisWrite(kw));

      test('', () => client.hisWrite(kw.id(), write)
      .then((_) {
        return new Future.delayed(new Duration(milliseconds: 200));
      })
      .then((_) => client.hisRead(kw.id(), "2010-06-07").then((HGrid read) {
        expect(read.numRows, equals(write.length));
        for (int i=0; i<read.numRows; ++i) {
          expect(read.row(i).get("ts"), equals(write[i].ts));
          expect(read.row(i).get("val"), equals(write[i].val));
        }
      })));
    });
  });
}

void verifyGridContains(HGrid g, String col, String val) => verifyGridContains2(g, col, new HStr(val));

void verifyGridContains2(HGrid g, String col, HVal val) {
  bool found = false;
  for (int i=0; i<g.numRows; ++i) {
    HVal x = g.row(i).get(col, checked:false);
    if (x != null && x == val) {
      found = true;
      break;
    }
  }

  expect(found, equals(true));
}